unit cDIBImageList;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: cDIBImageList.PAS, released August 28, 2000.

The Initial Developer of the Original Code is Peter Morris (pete@stuckindoors.com),
Portions created by Peter Morris are Copyright (C) 2000 Peter Morris.
All Rights Reserved.

Purpose of file:
This list may hold many images + their alpha masks.  All components should make use
of a DIBImageList rather than having the graphic data stored within the component.
(To save on application size)

Contributor(s):
None as yet


Last Modified: September 01, 2000
Current Version: 1.00

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org
or at http://www.stuckindoors.com/dib


Known Issues:
To be updated !
-----------------------------------------------------------------------------}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, cDIB;

type
  TDIBOperation = (doRemove, doChange);
  TDIBOperationEvent = procedure(Sender : TObject; Index : Integer; Operation : TDIBOperation) of object;

  TDIBImageList = class;

  TDIBImagesItem = Class(TCollectionItem)
  private
    FDIB                      : TMemoryDIB;
    FImportedFrom             : String;
    FDisplayName              : String;
    procedure ReadTemplateFilename(Reader : TReader);
    procedure SetDIB(const Value: TMemoryDIB);
    procedure WriteTemplateFilename(Writer : TWriter);
  protected
    procedure DefineProperties(Filer : TFiler); override;
    function GetDisplayName : String; override;
    procedure ImageChanged(Sender : TObject); virtual;
    procedure SetDisplayName(const Value : String); override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
  published
    property DIB              : TMemoryDIB
      read FDIB
      write SetDIB;
    property DisplayName;
  end;

  TDIBImages = class(TOwnedCollection)
  private
    { Private declarations }
    FOwner                    : TComponent;
    function  GetItem(Index : Integer) : TDIBImagesItem;
    procedure SetItem(Index : Integer; Value : TDIBImagesItem);
  protected
    procedure ImageChanged(Index : Integer; Operation : TDIBOperation); virtual;
    procedure Update(Item : TCollectionItem); override;
  public
    constructor Create(AOwner : TComponent);
    destructor Destroy; override;


    function Add : TDIBImagesItem;
    function AddTemplate(const GUID : String; const Index : Integer): TDIBImagesItem;

    property Items[Index : Integer]       : TDIBImagesItem
      read GetItem
      write SetItem; default;
  published
  end;

  TDIBImageLink = class(TPersistent)
  private
    FOwner                    : TComponent;
    FDIBImageList             : TDIBImageList;
    FDIBIndex                 : Integer;
    FOnImageChanged           : TDIBOperationEvent;
    procedure UnlinkNotification;
    procedure SetDIBImageList(const Value: TDIBImageList);
    procedure SetDIBImageIndex(const Value: Integer);
  protected
    procedure AssignTo(Dest : TPersistent); override;
    procedure ListChanged(Index : Integer; Operation : TDIBOperation); virtual;
  public
    constructor Create(AOwner : TComponent); virtual;
    destructor Destroy; override;

    function  GetImage(var ResultPic : TMemoryDIB) : Boolean;
    function  Valid : Boolean;

    property OnImageChanged   : TDIBOperationEvent
      read FOnImageChanged
      write FOnImageChanged;
  published
    property DIBImageList     : TDIBImageList
      read FDIBImageList
      write SetDIBImageList;
    property DIBIndex         : Integer
      read FDIBIndex
      write SetDIBImageIndex;
  end;

  TCustomDIBImageList = class(TComponent)
  private
    { Private declarations }
    FDIBImages                : TDIBImages;
    FLinkList                 : TList;
    FDuplicateDIB,
    FUniqueDIB                : TMemoryDIB;
  protected
    { Protected declarations }
    procedure AddLink(Link : TDIBImageLink); virtual;
    procedure ImageChanged(Sender : TObject; Index : Integer; Operation : TDIBOperation); virtual;
    procedure RemoveLink(Link : TDIBImageLink); virtual;

  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    function Get(Index : Integer) : TMemoryDIB;
    function GetUnique(Index : Integer) : TMemoryDIB;

    function GetImage(Index : Integer; var ResultPic : TMemoryDIB) : Boolean;
    function IsIndexValid(Index : Integer) : Boolean;
  published
    { Published declarations }
    property DIBImages        : TDIBImages
      read FDIBImages
      write FDIBImages;
  end;

  TDIBImageList = class(TCustomDIBImageList)
  published
  end;


procedure Register;

implementation
type
  THackPropertyDIB = class(TMemoryDIB);

procedure Register;
begin
  RegisterComponents('DIB', [TDIBImageList]);
end;

{ TDIBImagesItem }

constructor TDIBImagesItem.Create(Collection: TCollection);
begin
  inherited;
  FDIB := TMemoryDIB.Create(1,1);
  THackPropertyDIB(FDIB).OnChange := ImageChanged;
  FDisplayName := inherited GetDisplayName;
  FImportedFrom := '';
end;

procedure TDIBImagesItem.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('TemplateFilename',ReadTemplateFilename, WriteTemplateFilename, (FImportedFrom <> '')); 
end;

destructor TDIBImagesItem.Destroy;
var
  OldIndex                    : Integer;
  DIBImages                   : TDIBImages;
begin
  OldIndex := Self.Index;
  if Collection <> nil then DIBImages := TDIBImages(Collection);
  FDIB.Free;
  inherited;

  if DIBImages <> nil then
    DIBImages.ImageChanged(OldIndex, doRemove);
end;

function TDIBImagesItem.GetDisplayName: String;
begin
  Result := FDisplayName;
end;

procedure TDIBImagesItem.ImageChanged(Sender: TObject);
begin
  if Collection <> nil then
    TDIBImages(Collection).ImageChanged(Self.Index, doChange);
end;

procedure TDIBImagesItem.ReadTemplateFilename(Reader : TReader);
begin
  FImportedFrom := Reader.ReadString;
end;

procedure TDIBImagesItem.SetDIB(const Value: TMemoryDIB);
begin
  FDIB.Assign(Value);
end;

procedure TDIBImagesItem.SetDisplayName(const Value: String);
begin
  FDisplayName := Value;
  inherited;
end;

procedure TDIBImagesItem.WriteTemplateFilename(Writer : TWriter);
begin
  Writer.WriteString(FImportedFrom);
end;

{ TDIBImages }

function TDIBImages.Add: TDIBImagesItem;
begin
  Result := TDIBImagesItem(inherited Add);
end;

function TDIBImages.AddTemplate(const GUID: String;
  const Index: Integer): TDIBImagesItem;
var
  I,
  Position : Integer;
  SearchString : String;
begin
  Position := -1;
  SearchString := GUID + ':' + IntToStr(Index);
  
  for I:=0 to Count-1 do begin
    if Items[I].FImportedFrom = SearchString then begin
      Position := I;
      break;
    end;
  end;

  if Position > -1 then
    Result := Items[Position]
  else begin
    Result := Add;
    Result.FImportedFrom := GUID + ':' + IntToStr(Index);
  end;
end;

constructor TDIBImages.Create(AOwner: TComponent);
begin
  inherited Create(AOwner, TDIBImagesItem);
  FOwner := AOwner;
end;

destructor TDIBImages.Destroy;
begin
  inherited;
end;

function TDIBImages.GetItem(Index: Integer): TDIBImagesItem;
begin
  Result := TDIBImagesItem(inherited GetItem(Index));
end;

procedure TDIBImages.ImageChanged(Index : Integer; Operation : TDIBOperation);
begin
  if FOwner is TDIBImageList then
    TDIBImageList(FOwner).ImageChanged(Self, Index, Operation);
end;

procedure TDIBImages.SetItem(Index: Integer; Value: TDIBImagesItem);
begin
  inherited SetItem(Index, Value);
end;

procedure TDIBImages.Update(Item: TCollectionItem);
begin
  inherited Update(Item);
end;

{ TCustomDIBImageList }

procedure TCustomDIBImageList.AddLink(Link: TDIBImageLink);
var
  Index                       : Integer;
begin
  if Link = nil then exit;
  if FLinkList.IndexOf(Link) < 0 then FLinkList.Add(Link);
  if DIBImages <> nil then
    for Index := 0 to DIBImages.Count-1 do Link.ListChanged(Index, doChange);
end;

constructor TCustomDIBImageList.Create(AOwner: TComponent);
begin
  inherited;
  FDIBImages := TDIBImages.Create(Self);
  FDuplicateDIB := TMemoryDIB.Create(1,1);
  FUniqueDIB := TMemoryDIB.Create(1,1);
  FLinkList := TList.Create;
end;

destructor TCustomDIBImageList.Destroy;
var
  X                           : Integer;
begin
  For X:=FLinkList.Count-1 downto 0 do
    TDIBImageLink(FLinkList[X]).UnlinkNotification;
  FDIBImages.Free;
  FLinkList.Free;
  FUniqueDIB.Free;
  FDuplicateDIB.Free;
  inherited;
end;

function TCustomDIBImageList.Get(Index: Integer): TMemoryDIB;
begin
  Result := FDuplicateDIB;
  Result.PointDataAt(DIBImages.Items[Index].DIB);

end;

function TCustomDIBImageList.GetUnique(Index: Integer): TMemoryDIB;
begin
  FUniqueDIB.Assign(DIBImages.Items[Index].DIB);
  Result := FUniqueDIB;
end;

procedure TCustomDIBImageList.ImageChanged(Sender: TObject; Index: Integer; Operation : TDIBOperation);
var
  X                           : Integer;
begin
  for X:=0 to FLinkList.Count-1 do TDIBImageLink(FLinkList[X]).ListChanged(Index, Operation);
end;

function TCustomDIBImageList.GetImage(Index: Integer; var ResultPic: TMemoryDIB): Boolean;
var
  TheDIB                      : TMemoryDIB;
begin
  Result := False;
  if Self = nil then exit;
  if (Index < 0) or (Index >= DIBImages.Count) then exit;
  TheDIB := DIBImages.Items[Index].DIB;
  if not TheDIB.Valid then exit;

  ResultPic := Get(Index);
  Result := True;
end;

function TCustomDIBImageList.IsIndexValid(Index: Integer): Boolean;
begin
  Result := False;
  if Self = nil then exit;
  if (Index < 0) or (Index >= DIBImages.Count) then exit;
  Result := True;
end;

procedure TCustomDIBImageList.RemoveLink(Link: TDIBImageLink);
var
  Index                       : Integer;
begin
  if Link = nil then exit;

  Index := FLinkList.IndexOf(Link);
  if Index >=0 then begin
    FLinkList.Delete(Index);
//    if DIBImages <> nil then
//      for Index := 0 to DIBImages.Count-1 do
//        Link.ListChanged(Index, doRemove);
  end;
end;

{ TDIBImageLink }

procedure TDIBImageLink.AssignTo(Dest: TPersistent);
begin
  if Dest is TDIBImageLink then with TDIBImageLink(Dest) do begin
    DIBIndex := Self.DIBIndex;
    DIBImageList := Self.DIBImageList;
  end else
    inherited;
end;

constructor TDIBImageLink.Create(AOwner: TComponent);
begin
  inherited Create;
  FOwner := AOwner;
  FDIBIndex := -1;
end;

destructor TDIBImageLink.Destroy;
begin
  if DIBImageList <> nil then DIBImageList.RemoveLink(Self);
  inherited;
end;

function TDIBImageLink.GetImage(var ResultPic: TMemoryDIB): Boolean;
begin
  Result := False;
  if Self = nil then exit;
  Result := DIBImageList.GetImage(DIBIndex, ResultPic);
end;

procedure TDIBImageLink.ListChanged(Index : Integer; Operation : TDIBOperation);
begin
  if (Operation = doChange) and (Index <> DIBIndex) then exit;
  if (Operation = doRemove) and (Index > DIBIndex) then exit;

  if FOwner <> nil then
    if csDestroying in FOwner.ComponentState then exit;

  //If removing a list item which is < this one, we don't need an update,
  //we just change our index number
  if Operation = doRemove then
    if Index < DIBIndex then begin
      FDIBIndex := FDIBIndex -1;
      exit;
    end;

  if (Operation = doRemove) and (Index = DIBIndex) then FDIBIndex := -1;
  if Assigned(FOnImageChanged) then FOnImageChanged(Self, Index, Operation);
end;

procedure TDIBImageLink.SetDIBImageIndex(const Value: Integer);
var
  OldIndex                    : Integer;
begin
  OldIndex := FDIBIndex;
  FDIBIndex := Value;
  if (OldIndex <> Value) then ListChanged(Value, doChange);
end;

procedure TDIBImageLink.SetDIBImageList(const Value: TDIBImageList);
begin
  if FDIBImageList <> nil then FDIBImageList.RemoveLink(Self);
  FDIBImageList := Value;
  if FDIBImageList <> nil then FDIBImageList.AddLink(Self);
end;

procedure TDIBImageLink.UnlinkNotification;
begin
  DIBImageList := nil;
end;

function TDIBImageLink.Valid: Boolean;
begin
  Result := FDIBImageList.IsIndexValid(DIBIndex);
end;

end.
